/*
Module : PJNPOP3.cpp
Purpose: Implementation for a C++ class framework encapsulation of the POP3 protocol
Created: PJN / 04-05-1998
History: PJN / 27-06-1998 1. Fixed a potential buffer overflow problem in Delete
                          and Retrieve functions when a large message number was
                          specified.
                          2. Improve the ReadResponse code by a) passing the 
                          readability check onto the socket class and b) Sleeping
                          for 100 ms prior to looping around waiting for new 
                          response data
                          3. Classes are now fully Unicode compliant. Unicode
                          build configurations are also now included.
                          4. Now supports the TOP POP3 command which can be
                          issued using the GetHeader function.
         PJN / 04-01-1999 1. Properly UNICODE enabled the code
         PJN / 22-02-1999 1. Improved the reading of responses back from the server by implementing
                          a growable receive buffer
                          2. timeout is now 60 seconds when building for debug
                          3. Code now yields its time slice while waiting for a timeout
                          4. Made most functions virtual to help end user usage
         PJN / 25-03-1999 1. Fixed memory leak in the CPJNPOP3Connection::ReadReturnResponse function.
                          2. Now sleeps for 250 ms instead of yielding the time slice. This helps 
                          reduce CPU usage when waiting for data to arrive in the socket
         PJN / 15-06-1999 1. Added functions to return the message body, header or a particular
                          header field of a message
                          2. Tidied up some of the TRACE messages which the code generates
         PJN / 16-06-1999 1. Fixed a bug in the GetHeaderItem function which was causing a header
                          item which appeared at the beginning of the header fields to fail to be 
                          parsed incorrectly.
         PJN / 27-06-1999 1. Fixed a bug in the GetHeaderItem function when a header spanned 
                          multiple lines as is allowed by RFC 822
         PJN / 29-06-1999 1. Another improvement to GetHeaderItem. When will it end <g>. Originally 
                          this was reported as a bug but upon further investigation it turns out that
                          the message which was causing the problems had embedded tabs in the header. 
                          This is discouraged by the RFC which refers to mail headers (RFC 822). 
                          The code has been enhanced to handle this case. Thanks to Chris Bishop 
                          for spotting this.
                          2. Fix for a bug in GetHeaderItem where I accidently was using "=" instead of
                          "==". Thanks to Angelini Fabio for spotting this.
         PJN / 05-07-1999 1. Addition of the following functions:

                          i)   CPJNPOP3Message::GetReplyTo
                          ii)  CPJNPOP3Message::GetRawBody
                          iii) CPJNPOP3Message::GetSubject
                          iv)  CPJNPOP3Message::GetFrom
                          v)   CPJNPOP3Message::GetDate

                          2. GetHeaderItem function now uses case insensitive searching
                          3. GetHeaderItem now allows you to search for the "n'th" header of a specified type
         PJN / 24-08-1999 1. Fixed a bug whereby the function GetHeader was sometimes failing when it
                          was called when the message was retrieved using the "TOP" command.
         PJN / 27-03-2000 1. Fixed a problem where GetBody and GetRawBody will fail if you call it for a 
                          CPJNPOP3Message object that doesn't have a message in it (i.e m_pszMessage == NULL). 
                          This was previously causing a good old access violation in GetRawBody.
         PJN / 20-09-2000 1. Fixed a bug in CPJNPOP3Connection::UIDL when message id's were returned
                          with embedded dots in them. Thanks to Konstantin Vasserman for finding and resolving
                          this problem.
         PJN / 25-03-2001 1. Updated copyright information
         PJN / 27-07-2001 1. Added two methods namely: GetTo() and GetCC() to the CPJNPOP3Message class
         PJN / 11-08-2001 1. Fixed a bug in CPJNPOP3Connection::ReadResponse where determining the terminator
                          was failing when an embedded NULL was in the message.
         PJN / 27-09-2001 1. Fixed a bug in CPJNPOP3Connection::ReadResponse when handling disconnection errors
                          2. Improved the error handling in Delete, GetMessageSize & GetMessageID methods
         PJN / 29-09-2001 1. Further improved the error handling in CPJNPOP3Connection::ReadResponse
         PJN / 05-07-2003 1. Added copy constructors and operator= methods to the CPJNPOP3Message class. Thanks
                          to "Sten" for pointing out this omission.
                          2. Made destructors of all classes virtual.
         PJN / 21-07-2003 1. Class now preserves last command response across calls to Disconnect. This is 
                          especially important in calls to Connect when for example you get the POP3 error
                          "-Err. Can't lock the mailbox" or something like this when you call Connect. Thanks
                          to "Lev Elbert" for reporting this problem and providing the fix. 
         PJN / 03-11-2003 1. Simplified the code in CPJNPOP3Connection::ReadResponse. Thanks to Clarke Brunt for 
                          reporting this issue.
                          2. Reviewed the various trace statements in the code.
         PJN / 06-11-2003 1. Fixed a problem with an uninitialized "m_pszMessage" in the CPJNPOP3Message 
                          constructor. Thanks to Michael Gunlock for reporting this problem.
         PJN / 12-11-2003 1. Fixed a bug in the CPJNPOP3Message operator= method. Thanks to Henrik Grek for 
                          reporting this bug.
         PJN / 05-06-2004 1. Fixed a bug in CPJNPOP3Connection::ReadResponse, where the wrong parameters
                          were being null terminated. Thanks to "caowen" for this update.
         PJN / 04-07-2004 1. Added an IsConnected() method to the CPJNPOP3Connection class. Thanks to Alessandro 
                          Rosa for this addition.
                          2. Added two new helper functions, namely CPJNPOP3Message::GetEmailAddress and 
                          CPJNPOP3Message::GetEmailFriendlyName. Thanks to Alessandro Rosa for this suggestion.
         PJN / 09-08-2004 1. Fixed a number of compiler warnings when the code is compiled with VC.Net 2003
         PJN / 27-03-2005 1. Fixed a number of compiler warnings in the sample app when the code is compiled
                          with VC .NET 2003.
                          2. Timeout now defaults to 5 seconds in debug mode.
                          3. Added support for connecting via Socks4, Socks5 and HTTP proxies
                          4. Added support for POP3 via SSL. Thanks to Serhiy Pavlov and Alexey Kuznetsov for
                          suggesting this update. Please note that because I do not have access to a mail 
                          server which provides POP3 over SSL, you should consider this support preliminary
                          until I get feedback from users of this feature. The standard POP3 support is still
                          fully functional and tested of course.
         PJN / 09-04-2005 1. Updated the initialization code for OpenSSL in the sample app to avoid memory 
                          leaks. Thanks to Alexey Kuznetsov for reporting these issues. You should implement
                          the same initialization and cleanup code in your applications. For details please check
                          the code in "main.cpp" module which is enclosed in #ifndef POP3_NOSSL_SUPPORT / 
                          #endif sections.
                          2. "List" function is now public instead of protected. Thanks to Alexey Kuznetsov for 
                          reporting these issues. 
                          3. Addition of a simple "FindMessageID" which returns the Message index given a message
                          ID as returned via UIDL. Thanks to Alexey Kuznetsov for this addition. 
         PJN / 08-02-2006 1. Fixed a /Wp64 compile issue in CPJNPOP3Connection::FindMessageID. Thanks to Alexey 
                          Kuznetsov for reporting this issue.
                          2. Updated copyright details.
                          3. Renamed the classes to include a PJN prefix inline with my SMTP class.
                          4. Provided a PJNPOP3_EXT_CLASS to allow the class to be easily added to an extension dll
                          5. Changed the preprocessor macro to exclude SSL support to be CPJNPOP3_NOSSL. This is
                          to maintain consistency with my SMTP class.
                          6. Changed the default release mode timeout to 60 seconds and the default debug mode 
                          timeout to 90 seconds.
         PJN / 13-02-2006 1. Fixed a graceful disconnect issue in CPJNPOP3Connection::ReadResponse.
         PJN / 09-07-2006 1. Optimized CPJNPOP3Message constructor code.
                          2. Code now uses new C++ style casts rather than old style C casts where necessary.
                          3. Optimized CPJNPOP3Connection constructor code
                          4. All classes now uses exceptions to indicate errors. This means the whole area of 
                          handling errors in the code is a whole lot simpler. You should carefully review your 
                          code as a lot of the return values from methods (especially CPJNPOP3Connection) are now void 
                          and will throw CPJNPOP3Exception's.
                          5. General rework of the class to bring it in line with the author's CPJNSMTPConnection code
                          6. Update the code in the sample app to TRACE out what it is doing
                          7. Updated the code to clean compile on VC 2005
         PJN / 15-11-2006 1. Minor update to the sample app to allow it to clean compile on VC 2005
                          2. Optimized ReadResponse method so that if an error occurs while reading the "RETR" 
                          response, only the actual command text is set into m_sLastCommandResponse instead of the 
                          whole message. This avoids a great performance hit on Unicode projects because of the
                          m_sLastCommandResponse CString nature (ANSI -> UNICODE conversion). Thanks to Alexey 
                          Kuznetsov for reporting this issue.
         PJN / 21-07-2007 1. Updated copyright details
                          2. CPJNPOP3Connection::Connect, Delete, Statistics, List, UIDL, Reset, Noop and GetMessageHeader
                          methods now catch CWSocketException exceptions and rethrow them as CPJNPOP3Exception exceptions. 
                          This allows client code to only expect CPJNPOP3Exception exceptions to be thrown from the class. 
                          Thanks to Alexey Kuznetsov for reporting this issue.
         PJN / 31-12-2007 1. Updated sample app to clean compile on VC 2008
                          2. CPJNPOP3Exception::GetErrorMessage now uses the FORMAT_MESSAGE_IGNORE_INSERTS flag. For more 
                          information please see Raymond Chen's blog at 
                          http://blogs.msdn.com/oldnewthing/archive/2007/11/28/6564257.aspx. Thanks to Alexey Kuznetsov for 
                          reporting this issue.
                          3. CPJNPOP3Exception::GetErrorMessage now uses Checked::tcsncpy_s if compiled using VC 2005 or 
                          later.
                          4. Sample app shipped with download is now a VC 2005 solution
         PJN / 18-03-2008 1. Updated copyright details
                          2. The code now only supports VC 2005 or later.
                          3. Simplified the response parsing considerably through the usage of CStringA
                          4. Fixed logic which terminates the parsing loop in ReadListResponse
                          5. Improved the robustness of the ReadUIDLResponse parsing logic
                          6. Now includes support for APOP authentication. Thanks to "goodsiwen" for prompting this update.
         PJN / 18-05-2008 1. GetHeaderItem now uses a CStringA parameter instead of CString. Thanks to Alexey Kuznetsov for
                          reporting this issue.
         PJN / 31-05-2008 1. Code now compiles cleanly using Code Analysis (/analyze)
         PJN / 14-06-2008 1. Updated code to compile correctly using _ATL_CSTRING_EXPLICIT_CONSTRUCTORS define
         PJN / 04-05-2009 1. Updated copyright details.
                          2. Updated the sample app's project settings to more modern default values.
                          3. Removed the parsing methods from the CPJNPOP3Message class. For comprehensive support for 
                          message parsing please take a look at CDO for Windows 2000.
                          4. Extensively reworked the parsing code to use CByteArray derived classes to hold parsed 
                          responses instead of CStringA instances. This handles the unusual but legal situation where a 
                          message contains embedded nulls. Thanks to Sergey Lysenko for reporting this issue.
                          5. Fixed a bug in ReadUIDLResponse when parsing the message ID
                          6. The code in Connect now correctly validates that a timestamp has been provided if doing APOP
                          authentication
         PJN / 22-12-2009 1. Optimized the ReadResponse logic to reduce the amount of searching required to find the terminator 
                          in the received buffer. This helps avoid excessive CPU utilization when we receive large responses.
                          Thanks to Alexey Rabets for reporting this issue.
                          2. The size of the read buffer used in the ReadResponse method can now be specified at runtime via 
                          SetReadBufferSize().
                          3. Updated the documentation to fix various out of date details in the API reference section.
         PJN / 13-02-2011 1. Updated copyright details
                          2. Updated sample app to clean compile on VC 2010
                          3. Updated code to support latest SSL and Sockets class from the author. This means that the code now supports IPv6 POP3 servers
                          4. The sample app is now linked against the latest OpenSSL v1.0.0c dlls
                          5. Set/GetBoundAddress have been renamed Set/GetBindAddress for consistency with the sockets class
         PJN / 09-02-2014 1. Updated the code to clean compile on VC 2012 & VC 2013.
                          2. Updated copyright details
                          3. Removed all the proxy connection methods as they cannot be easily supported / tested by the author.
                          4. The sample app is now linked against the latest OpenSSL v1.0.1f dlls.
                          5. The CPJNPOP3Connection class now provides Set/GetSSLProtocol methods. This allows client code to specify the exact flavour 
                          of SSL which the code should speak. Supported protocols are SSL v2/v3, TLS v1.0, TLS v1.1, TLS v1.2 and DTLS v1.0. The default 
                          is to use TLS v1.0.
                          6. CPJNPOP3Connection::_Close now provides a bGracefully parameter
                          7. Made more methods virtual to facilitate further client customization
         PJN / 13-04-2014 1. The sample app is now linked against the latest OpenSSL v1.0.1g dlls. This version is the patched version of OpenSSL which does not
                          suffer from the Heartbleed bug.
                          2. Please note that by default OpenSSL does not do host name validation. The sample app provided with the PJNPOP3 code also does not do
                          host name validation. This means that as it stands the sample app is vulnerable to man in the middle attacks if you use SSL/TLS to connect
                          to a POP3 server. For further information and sample code which you should incorporate into your real SMTP client applications, please see
                          http://wiki.openssl.org/index.php/Hostname_validation, https://github.com/iSECPartners/ssl-conservatory and 
                          http://archives.seul.org/libevent/users/Feb-2013/msg00043.html.
         PJN / 23-12-2014 1. Updated the code to use the author's SSLWrappers classes (http://www.naughter.com/sslwrappers.html) to provide the SSL functionality 
                          for PJNPOP3 instead of OpenSSL. Please note that the SSLWrappers classes itself depends on the author's CryptoWrappers classes 
                          (http://www.naughter.com/cryptowrappers.html) also. You will need to download both of these libraries and copy their modules into the 
                          PJNPOP3 directory. Also note that the SSLWrappers and CryptoWrapper classes are only supported on VC 2013 or later and this means that 
                          the PJNPOP3 SSL support is only supported when you compile with VC 2013 or later.
                          2. With the change to using SSLWrappers, the code can now support additional flavours of SSL protocol. They are: SSL v2 on its own, 
                          SSL v3 on its own, The OS SChannel SSL default and AnyTLS. Please see the CPJNPOP3Connection::SSLProtocol enum for more details. The 
                          sample app has been updated to allow these values to be used and the default enum value is now "OSDefault".
                          3. The default timeout set in the CPJNPOP3Connection constructor is now 60 seconds for both debug and release builds
                          4. The sample app shipped with the source code is now Visual Studio 2008 and as of this release the code is only supported on Visual 
                          Studio 2008 and later
         PJN / 12-06-2016 1. Updated the code to compile cleanly on VC 2015
                          2. Updated copyright details.
                          3. Added SAL annotations to all the code
                          4. Reworked the classes to optionally compile without MFC. By default the classes now use STL classes and idioms but if you define 
                          CWSOCKET_MFC_EXTENSTIONS the classes will revert back to the MFC behaviour.
         PJN / 22-07-2016 1. Updated code to use new preprocessor define of CWSOCKET_MFC_EXTENSIONS instead of CWSOCKET_MFC_EXTENSTIONS
                          2. Minor update to the sample project settings to remove _MBCS preprocessor macro when compiling for Unicode.
         PJN / 19-08-2016 1. Updated the code to work with the latest version of the MD5 classes of the author which can operate in non MFC mode.
                          2. Reworked CPJNPOP3Connection class to be inherited from SSLWrappers::CSocket. This allows for easier overloading of the various 
                          SSLWrappers class framework virtual functions
                          3. Reworked the CPJNPOP3Exception::GetErrorMessage to not bother putting the last response into the return error message text.
                          4. Tested the classes support for the APOP command using a free trial version of the MDaemon Messaging server
         PJN / 26-03-2017 1. Updated copyright details.
                          2. Updated the code to clean compile on VC 2017
         PJN / 29-04-2017 1. Updated the code to compile cleanly using /permissive-.
         PJN / 12-09-2017 1. Replaced CString::operator LPC*STR() calls throughout the codebase with CString::GetString calls
         PJN / 18-12-2017 1. Updated the code to compile cleanly when _ATL_NO_AUTOMATIC_NAMESPACE is defined.
                          2. Removed DTLSv1 as a supported protocol from SSLProtocol enum
         PJN / 25-06-2018 1. Updated copyright details.
                          2. Fixed a number of C++ core guidelines compiler warnings. These changes mean that
                          the code will now only compile on VC 2017 or later.
                          3. CPJNPOP3String is now always defined as std::string or std::wstring and no longer supports CString
         PJN / 04-11-2018 1. Fixed some further C++ core guidelines compiler warnings.
         PJN / 21-04-2019 1. Updated copyright details
                          2. Updated the code to clean compile on VC 2019
                          3. Removed all dependencies on now defunct IDS_PJNPOP3* string resources
         PJN / 02-10-2019 1. Fixed a number of compiler warnings when the code is compiled with VS 2019 Preview
         PJN / 27-12-2019 1. Fixed various Clang-Tidy static code analysis warnings in the code.
         PJN / 07-01-2020 1. Updated copyright details.
                          2. Fixed some missing catch statements in the Connect method. Thanks to Martin Richter for 
                          reporting this issue.
                          3. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 04-05-2020 1. Fixed more Clang-Tidy static code analysis warnings in the code.
                          2. Updated copyright details.
         PJN / 23-04-2021 1. Updated the code to work with the latest version of the hash classes of the author.
                          2. Updated copyright details.
         PJN / 08-01-2022 1. Updated copyright details.
                          2. Updated the code and sample app to support TLS v1.3
         PJN / 18-04-2022 1. Updated the code to use C++ uniform initialization for all variable declarations
                          2. Reworked CPJNPOP3Buffer::Find to use std::search.
         PJN / 04-12-2022 1. Fixed an issue in CPJNPOP3Connection::ReadSSLResponse when handling SEC_I_RENEGOTIATE return
                          values from GetEncryptedMessage.
         PJN / 27-05-2023 1. Updated copyright details.
                          2. Updated module to indicate that it needs to be compiled using /std:c++17. Thanks to Martin Richter
                          for reporting this issue.

Copyright (c) 1998 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Includes /////////////////////////////////////////////////

#include "stdafx.h"
#include "PJNPOP3.h"
#include "PJNHash.h" //If you get a compilation error about this missing header file, then you need to download my Hash wrapper classes from http://www.naughter.com/smtp.html


//////////////////// Macros / Defines /////////////////////////////////////////

#ifdef _DEBUG
#define new DEBUG_NEW
#endif //#ifdef _DEBUG


//////////////////// Implementation ///////////////////////////////////////////

CPJNPOP3String CPJNPOP3Buffer::AsString() const
{
  //What will be the return value from this function
  CPJNPOP3String sResponse;

  const size_t nSize{size()};
  sResponse.resize(nSize);
  if (nSize)
  {
#pragma warning(suppress: 26446 26429)
    TCHAR* pszResponse{sResponse.data()};
    int nOutputIndex{0};
    for (size_t i{0}; i<nSize; i++)
    {
      //Lets simple skip over any embedded nulls
#pragma warning(suppress: 26446)
      const BYTE byData{operator[](i)};
      if (byData)
      {
#pragma warning(suppress: 26481)
        pszResponse[nOutputIndex] = byData;
        ++nOutputIndex;
      }
    }
#pragma warning(suppress: 26481)
    pszResponse[nOutputIndex] = _T('\0');
  }

  return sResponse;
}

CPJNPOP3Buffer CPJNPOP3Buffer::Left(_In_ INT_PTR nCount) const
{
  CPJNPOP3Buffer buffer;
  if (nCount)
#pragma warning(suppress: 26444)
    buffer.insert(buffer.begin(), begin(), begin() + nCount);
  return buffer;
}

#pragma warning(suppress: 26440)
int CPJNPOP3Buffer::Find(_In_z_ LPCSTR pszToFind, _In_ int nStart) const
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSUME(pszToFind != nullptr);

#pragma warning(suppress: 26481 26490)
  auto iter = std::search(cbegin() + nStart, cend(), reinterpret_cast<const BYTE*>(pszToFind), reinterpret_cast<const BYTE*>(pszToFind) + strlen(pszToFind));
  if (iter == cend())
    return -1;
  else
#pragma warning(suppress: 26472)
    return static_cast<int>(iter - cbegin());
}


CPJNPOP3Exception::CPJNPOP3Exception(_In_ HRESULT hr, _In_ CPJNPOP3String sLastResponse) noexcept : m_hr{hr},
                                                                                                    m_sLastResponse{std::move(sLastResponse)}
{
}

CPJNPOP3Exception::CPJNPOP3Exception(_In_ DWORD dwError, _In_ DWORD dwFacility, _In_ CPJNPOP3String sLastResponse) noexcept : m_hr{MAKE_HRESULT(SEVERITY_ERROR, dwFacility, dwError)},
                                                                                                                              m_sLastResponse{std::move(sLastResponse)}
{
}

#pragma warning(suppress: 26429)
bool CPJNPOP3Exception::GetErrorMessage(_Out_writes_z_(nMaxError) LPTSTR lpszError, _In_ UINT nMaxError, _Out_opt_ PUINT pnHelpContext)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(lpszError != nullptr);
#pragma warning(suppress: 26477)
  ATLASSERT(nMaxError != 0);
#pragma warning(suppress: 26481)
  lpszError[0] = _T('\0');

  if (pnHelpContext != nullptr)
    *pnHelpContext = 0;

  //What will be the return value from this function (assume the worst)
  bool bSuccess{false};

  if (HRESULT_FACILITY(m_hr) == FACILITY_ITF)
  {
    //Simply load up the string from the string table
    CPJNPOP3String sError;
    ATL::CAtlString sResource;
    bSuccess = sResource.LoadString(GetModuleHandle(nullptr), HRESULT_CODE(m_hr));
    if (bSuccess)
    {
      sError = sResource;
      ATL::Checked::tcsncpy_s(lpszError, nMaxError, sError.c_str(), _TRUNCATE);
    }
  }
  else
  {
    LPTSTR lpBuffer{nullptr};
    const DWORD dwReturn{FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
                                       nullptr, HRESULT_CODE(m_hr), MAKELANGID(LANG_NEUTRAL, SUBLANG_SYS_DEFAULT),
#pragma warning(suppress: 26490)
                                         reinterpret_cast<LPTSTR>(&lpBuffer), 0, nullptr)};
    if (dwReturn == 0)
      *lpszError = _T('\0');
    else
    {
      bSuccess = true;
      ATL::Checked::tcsncpy_s(lpszError, nMaxError, lpBuffer, _TRUNCATE);
      LocalFree(lpBuffer);
    }
  }

  return bSuccess;
}


CPJNPOP3Connection::CPJNPOP3Connection() noexcept : m_bSSL{false},
                                                    m_bConnected{false},
                                                    m_dwTimeout{60000},
                                                    m_dwReadBufferSize{1024},
                                                    m_nNumberOfMails{0},
                                                    m_bListRetrieved{false},
                                                    m_bStatRetrieved{false},
                                                    m_bUIDLRetrieved{false},
                                                    m_SSLProtocol{SSLProtocol::OSDefault}
{
}

CPJNPOP3Connection::~CPJNPOP3Connection()
{
  if (m_bConnected)
  {
    //Deliberately handle exceptions here, so that the destructor does not throw exceptions
    try
    {
#pragma warning(suppress: 26447)
      Disconnect(true);
    }
    catch(CPJNPOP3Exception&)
    {
    }
  }
}

void CPJNPOP3Connection::ThrowPJNPOP3Exception(_In_ DWORD dwError, _In_ DWORD dwFacility, _In_ const CPJNPOP3String& sLastResponse)
{
  if (dwError == 0)
    dwError = ::WSAGetLastError();

  CPJNPOP3Exception e{dwError, dwFacility, sLastResponse};
  ATLTRACE(_T("Warning: throwing CPJNPOP3Exception for error %08X\n"), e.m_hr);
  throw e;
}

void CPJNPOP3Connection::ThrowPJNPOP3Exception(_In_ HRESULT hr, _In_ const CPJNPOP3String& sLastResponse)
{
  CPJNPOP3Exception e{hr, sLastResponse};
  ATLTRACE(_T("Warning: throwing CPJNPOP3Exception for error %08X\n"), e.m_hr);
  throw e;
}

void CPJNPOP3Connection::_Send(_In_reads_bytes_(nBuf) const void* pBuffer, _In_ int nBuf)
{
#ifndef CPJNPOP3_NOSSL
  if (GetSocket() != INVALID_SOCKET)
  {
    const SECURITY_STATUS ss{SendEncrypted(static_cast<const BYTE*>(pBuffer), nBuf)};
    if (ss != SEC_E_OK)
      ThrowPJNPOP3Exception(ss);
  }
  else
#endif //#ifndef CPJNSMTP_NOSSL
    m_Socket.Send(pBuffer, nBuf);
}

#ifndef CPJNPOP3_NOSSL
void CPJNPOP3Connection::CreateSSLSocket()
{
  //Acquire the credentials
  if (m_SSLCredentials.get())
  {
    if (m_SSLCredentials->ValidHandle())
      m_SSLCredentials->Free();
    m_SSLCredentials.reset();
  }
  if ((m_SSLProtocol == SSLProtocol::AnyTLSUpTo1_3) || (m_SSLProtocol == SSLProtocol::TLSv1_3))
    m_SSLCredentials = std::make_unique<SSLWrappers::CCachedCredentials>(true);
  else
    m_SSLCredentials = std::make_unique<SSLWrappers::CCachedCredentials>();
  TLS_PARAMETERS TLSParams{};
  switch (m_SSLProtocol)
  {
    case SSLProtocol::SSLv2:
    {
       m_SSLCredentials->m_sslCredentials.cSupportedAlgs = SP_PROT_SSL2_CLIENT;
       break;
    }
    case SSLProtocol::SSLv3:
    {
       m_SSLCredentials->m_sslCredentials.cSupportedAlgs = SP_PROT_SSL3_CLIENT;
       break;
    }
    case SSLProtocol::SSLv2orv3:
    {
       m_SSLCredentials->m_sslCredentials.cSupportedAlgs = SP_PROT_SSL2_CLIENT | SP_PROT_SSL3_CLIENT;
       break;
    }
    case SSLProtocol::TLSv1:
    {
       m_SSLCredentials->m_sslCredentials.cSupportedAlgs = SP_PROT_TLS1_0_CLIENT;
       break;
    }
    case SSLProtocol::TLSv1_1:
    {
       m_SSLCredentials->m_sslCredentials.cSupportedAlgs = SP_PROT_TLS1_1_CLIENT;
       break;
    }
    case SSLProtocol::TLSv1_2:
    {
       m_SSLCredentials->m_sslCredentials.cSupportedAlgs = SP_PROT_TLS1_2_CLIENT;
       break;
    }
    case SSLProtocol::OSDefault:
    {
       m_SSLCredentials->m_sslCredentials.cSupportedAlgs = 0;
       break;
    }
    case SSLProtocol::AnyTLSUpTo1_2:
    {
      m_SSLCredentials->m_sslCredentials.cSupportedAlgs = SP_PROT_TLS1_0_CLIENT | SP_PROT_TLS1_1_CLIENT | SP_PROT_TLS1_2_CLIENT;
      break;
    }
    case SSLProtocol::TLSv1_3:
    {
      m_SSLCredentials->m_sslCredentials2.cTlsParameters = 1;
      m_SSLCredentials->m_sslCredentials2.pTlsParameters = &TLSParams;
      TLSParams.grbitDisabledProtocols = SP_PROT_SSL2_CLIENT | SP_PROT_PCT1_CLIENT | SP_PROT_SSL3_CLIENT |
                                         SP_PROT_TLS1_0_CLIENT | SP_PROT_TLS1_1_CLIENT | SP_PROT_TLS1_2_CLIENT;
      break;
    }
    case SSLProtocol::AnyTLSUpTo1_3:
    {
      m_SSLCredentials->m_sslCredentials2.cTlsParameters = 1;
      m_SSLCredentials->m_sslCredentials2.pTlsParameters = &TLSParams;
      TLSParams.grbitDisabledProtocols = SP_PROT_SSL2_CLIENT | SP_PROT_PCT1_CLIENT | SP_PROT_SSL3_CLIENT;
      break;
    }
    default:
    {
#pragma warning(suppress: 26477)
      ATLASSERT(false);
      break;
    }
  }
  const SECURITY_STATUS ss{m_SSLCredentials->AcquireClient()};
  if (ss != SEC_E_OK)
    ThrowPJNPOP3Exception(ss);

  //Setup the SSLWrappers::CSocket instance
  if (ValidHandle())
    SSLWrappers::CSocket::Delete();
  SetCachedCredentials(m_SSLCredentials.get());
  SetReadTimeout(m_dwTimeout);
  SetWriteTimeout(m_dwTimeout);
  Attach(m_Socket);
}
#endif //#ifndef CPJNPOP3_NOSSL

#ifndef CPJNPOP3_NOSSL
void CPJNPOP3Connection::Connect(_In_z_ LPCTSTR pszHostName, _In_z_ LPCTSTR pszUsername, _In_z_ LPCTSTR pszPassword, _In_ int nPort, _In_ AuthenticationMethod am, _In_ bool bSSL)
#else
void CPJNPOP3Connection::Connect(_In_z_ LPCTSTR pszHostName, _In_z_ LPCTSTR pszUsername, _In_z_ LPCTSTR pszPassword, _In_ int nPort, _In_ AuthenticationMethod am)
#endif //#ifndef CPJNPOP3_NOSSL
{
  //Validate our parameters
#ifndef CPJNPOP3_NOSSL
  m_bSSL = bSSL;
#else
  m_bSSL = false;
#endif //#ifndef CPJNPOP3_NOSSL
  m_sHostName = pszHostName;

  try
  {
    m_Socket.SetBindAddress(m_sBindAddress.c_str());
    m_Socket.CreateAndConnect(pszHostName, nPort);

  #ifndef CPJNPOP3_NOSSL
    if (m_bSSL)
    {
      CreateSSLSocket();
      const SECURITY_STATUS ss{SSLConnect(pszHostName)};
      if (ss != SEC_E_OK)
      {
        //Disconnect before we throw the exception
        try
        {
          Disconnect(true);
        }
        catch (CPJNPOP3Exception&)
        {
        }
        ThrowPJNPOP3Exception(ss);
      }
    }
  #endif //#ifndef CPJNPOP3_NOSSL
  }

#ifdef CWSOCKET_MFC_EXTENSIONS
#pragma warning(suppress: 26429)
  catch(CWSocketException* pEx)
  {
    //Disconnect before we rethrow the exception
    try
    {
      Disconnect(true);
    }
    catch(CPJNPOP3Exception&)
    {
    }

#pragma warning(suppress: 26472)
    const DWORD dwError{static_cast<DWORD>(pEx->m_nError)};
    pEx->Delete();
    ThrowPJNPOP3Exception(dwError, FACILITY_WIN32);
  }
#else
#pragma warning(suppress: 26496)
  catch(CWSocketException& e)
  {
    //Disconnect before we rethrow the exception
    try
    {
      Disconnect(true);
    }
    catch(CPJNPOP3Exception&)
    {
    }

    //rethrow the exception
    ThrowPJNPOP3Exception(e.m_nError, FACILITY_WIN32);
  }
#endif //#ifdef CWSOCKET_MFC_EXTENSIONS

  //At this point, we're now connected
  m_bConnected = true;

  try
  {
    //check the response
    CPJNPOP3String sTimeStampBanner;
#pragma warning(suppress: 26489)
    CPJNPOP3String sResponse;
    if (!ReadCommandResponse(sResponse))
      ThrowPJNPOP3Exception(ERROR_UNEXP_NET_ERR, FACILITY_WIN32, m_sLastCommandResponse);
    else
    {
      //Hive away the timestamp banner
      const CPJNPOP3String::size_type nFind1{sResponse.find(_T("<"))};
      const CPJNPOP3String::size_type nFind2{sResponse.find(_T(">"))};
      if ((nFind2 != CPJNPOP3String::npos) && (nFind1 != CPJNPOP3String::npos) && (nFind2 > nFind1))
        sTimeStampBanner = sResponse.substr(nFind1, nFind2 - nFind1 + 1);
    }

    //Form ascii versions of the username and password parameters
    ATL::CT2A sAsciiUsername{pszUsername};
    ATL::CT2A sAsciiPassword{pszPassword};

    //Issue the correct sequence of commands depending on the authentication scheme 
    if (am == AuthenticationMethod::AUTH_PLAIN)
    {
      //Send the POP3 username and check the response
      std::ostringstream ss;
      ss << "USER " << sAsciiUsername.operator LPSTR() << "\r\n";
      std::string sRequest{ss.str()};
#pragma warning(suppress: 26472)
      _Send(sRequest.c_str(), static_cast<int>(sRequest.size()));
      if (!ReadCommandResponse())
        ThrowPJNPOP3Exception(ERROR_UNEXP_NET_ERR, FACILITY_WIN32, m_sLastCommandResponse);

      //Send the POP3 password and check the response
      std::ostringstream ss2;
      ss2 << "PASS " << sAsciiPassword.operator LPSTR() << "\r\n";
      sRequest = ss2.str();
#pragma warning(suppress: 26472)
      _Send(sRequest.c_str(), static_cast<int>(sRequest.size()));
      if (!ReadCommandResponse())
        ThrowPJNPOP3Exception(ERROR_UNEXP_NET_ERR, FACILITY_WIN32, m_sLastCommandResponse);
    }
    else
    {
#pragma warning(suppress: 26477)
      ATLASSERT(am == AuthenticationMethod::AUTH_APOP);

      //Check that the server provided us a timestamp to hash
      if (sTimeStampBanner.length() == 0)
        ThrowPJNPOP3Exception(ERROR_UNEXP_NET_ERR, FACILITY_WIN32, m_sLastCommandResponse);

      //Form the string which we need to MD5 hash
      std::string sToHash{ATL::CT2A{sTimeStampBanner.c_str()}};
      sToHash += sAsciiPassword;

      //Calculate the MD5 hash
      CPJNHashResult hash;
      CPJNHash md5{BCRYPT_MD5_ALGORITHM};
#pragma warning(suppress: 26490 26472)
      const NTSTATUS status{md5.Hash(reinterpret_cast<const BYTE*>(sToHash.c_str()), static_cast<DWORD>(sToHash.length()), hash)};
      if (!BCRYPT_SUCCESS(status))
        ThrowPJNPOP3Exception(status);

      //Issue the APOP command and check the response
      ATL::CT2A sAPOP{hash.Format().c_str()};
      std::ostringstream ss;
      ss << "APOP " << sAsciiUsername.operator LPSTR() << " " << sAPOP.operator LPSTR() << "\r\n";
      const std::string sRequest{ss.str()};
#pragma warning(suppress: 26472)
      _Send(sRequest.c_str(), static_cast<int>(sRequest.length()));
      if (!ReadCommandResponse())
        ThrowPJNPOP3Exception(ERROR_UNEXP_NET_ERR, FACILITY_WIN32, m_sLastCommandResponse);
    }
  }
#ifdef CWSOCKET_MFC_EXTENSIONS
#pragma warning(suppress: 26429)
  catch(CWSocketException* pEx)
  {
    //Disconnect before we rethrow the exception
    try
    {
      Disconnect(true);
    }
    catch (CPJNPOP3Exception&)
    {
    }

#pragma warning(suppress: 26472)
    const DWORD dwError{static_cast<DWORD>(pEx->m_nError)};
    pEx->Delete();
    ThrowPJNPOP3Exception(dwError, FACILITY_WIN32);
  }
#else
#pragma warning(suppress: 26496)
  catch(CWSocketException& e)
  {
    //Disconnect before we rethrow the exception
    try
    {
      Disconnect(true);
    }
    catch(CPJNPOP3Exception&)
    {
    }

    ThrowPJNPOP3Exception(e.m_nError, FACILITY_WIN32);
  }
#endif //#ifdef CWSOCKET_MFC_EXTENSIONS
#pragma warning(suppress: 26496)
  catch(CPJNPOP3Exception& e)
  {
    //Disconnect before we rethrow the exception
    try
    {
      Disconnect(true);
    }
    catch(CPJNPOP3Exception&)
    {
    }

    //rethrow the exception
    ThrowPJNPOP3Exception(e.m_hr, e.m_sLastResponse);
  }
}

void CPJNPOP3Connection::Disconnect(_In_ bool bGracefully)
{
  //disconnect from the POP3 server if connected 
  HRESULT hr{S_OK};
  CPJNPOP3String sLastResponse;
  if (m_bConnected)
  {
    if (bGracefully)
    {
      static const char* sBuf{"QUIT\r\n"};
      try
      {
#pragma warning(suppress: 26472)
        _Send(sBuf, static_cast<int>(strlen(sBuf)));
      }
    #ifdef CWSOCKET_MFC_EXTENSIONS
#pragma warning(suppress: 26429 26462)
      catch(CWSocketException* pEx)
      {
        hr = MAKE_HRESULT(SEVERITY_ERROR, pEx->m_nError, FACILITY_WIN32);
        pEx->Delete();
      }
    #else
#pragma warning(suppress: 26496)
      catch(CWSocketException& e)
      {
        hr = MAKE_HRESULT(SEVERITY_ERROR, e.m_nError, FACILITY_WIN32);
      }
    #endif //#ifdef CWSOCKET_MFC_EXTENSIONS

      //Check the response
      if (!ReadCommandResponse())
      {
        hr = MAKE_HRESULT(SEVERITY_ERROR, ERROR_UNEXP_NET_ERR, FACILITY_WIN32);
        sLastResponse = GetLastCommandResponse();
      }
    }

    //Reset the connected state
    m_bConnected = false;
  }

  //free up our socket
#ifndef CPJNPOP3_NOSSL
  if (m_SSLCredentials.get())
  {
    if (m_SSLCredentials->ValidHandle())
      m_SSLCredentials->Free();
    m_SSLCredentials.reset();
  }
  if (ValidHandle())
    SSLWrappers::CSocket::Delete();
  Detach();
#endif //#ifdef CPJNPOP3_NOSSL
  m_Socket.Close();

  //Reset all our member variables
  m_bListRetrieved = false;
  m_bStatRetrieved = false;
  m_bUIDLRetrieved = false;
  m_msgSizes.clear();
  m_msgIDs.clear();

  //Should we throw an exception
  if (FAILED(hr))
    ThrowPJNPOP3Exception(hr, sLastResponse);
}

#pragma warning(suppress: 26434)
void CPJNPOP3Connection::Delete(_In_ int nMsg)
{
  //Must be connected to perform a delete
#pragma warning(suppress: 26477)
  ATLASSERT(m_bConnected);

  //if we haven't executed the LIST command then do it now
  if (!m_bListRetrieved)
    List();

  //Send the DELE command along with the message ID
  std::ostringstream ss;
  ss << "DELE " << nMsg << "\r\n";
  const std::string sRequest{ss.str()};
  try
  {
#pragma warning(suppress: 26472)
    _Send(sRequest.c_str(), static_cast<int>(sRequest.length()));
  }
#ifdef CWSOCKET_MFC_EXTENSIONS
#pragma warning(suppress: 26429 26462)
  catch(CWSocketException* pEx)
  {
#pragma warning(suppress: 26472)
    const DWORD dwError{static_cast<DWORD>(pEx->m_nError)};
    pEx->Delete();
    ThrowPJNPOP3Exception(dwError, FACILITY_WIN32);
  }
#else
#pragma warning(suppress: 26496)
  catch(CWSocketException& e)
  {
    ThrowPJNPOP3Exception(e.m_nError, FACILITY_WIN32);
  }
#endif //#ifdef CWSOCKET_MFC_EXTENSIONS

  if (!ReadCommandResponse())
    ThrowPJNPOP3Exception(ERROR_UNEXP_NET_ERR, FACILITY_WIN32, m_sLastCommandResponse);
}

void CPJNPOP3Connection::Statistics(_Out_ int& nNumberOfMails, _Out_ int& nTotalMailSize)
{
  //Must be connected to perform a "STAT"
#pragma warning(suppress: 26477)
  ATLASSERT(m_bConnected);

  //Send the STAT command
  static const char* sBuf{"STAT\r\n"};
  try
  {
#pragma warning(suppress: 26472)
    _Send(sBuf, static_cast<int>(strlen(sBuf)));
  }
#ifdef CWSOCKET_MFC_EXTENSIONS
#pragma warning(suppress: 26429 26462)
  catch(CWSocketException* pEx)
  {
#pragma warning(suppress: 26472)
    const DWORD dwError{static_cast<DWORD>(pEx->m_nError)};
    pEx->Delete();
    ThrowPJNPOP3Exception(dwError, FACILITY_WIN32);
  }
#else
#pragma warning(suppress: 26496)
  catch(CWSocketException& e)
  {
    ThrowPJNPOP3Exception(e.m_nError, FACILITY_WIN32);
  }
#endif //#ifdef CWSOCKET_MFC_EXTENSIONS

  ReadStatResponse(nNumberOfMails, nTotalMailSize);
  m_bStatRetrieved = true;
}

DWORD CPJNPOP3Connection::GetMessageSize(_In_ int nMsg)
{
  //if we haven't executed the LIST command then do it now
  if (!m_bListRetrieved)
    List();

  //retrieve the size from the message size array
#pragma warning(suppress: 26451 26446)
  return m_msgSizes[nMsg - 1];
}

CPJNPOP3String CPJNPOP3Connection::GetMessageID(_In_ int nMsg)
{
  //if we haven't executed the UIDL command then do it now
  if (!m_bUIDLRetrieved)
    UIDL();

  //retrieve the ID from the array
#pragma warning(suppress: 26472 26446)
  return m_msgIDs[static_cast<size_t>(nMsg) - 1];
}

int CPJNPOP3Connection::FindMessageID(_In_z_ LPCTSTR pszID)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(_tcslen(pszID));

  //if we haven't executed the UIDL command then do it now
  if (!m_bUIDLRetrieved)
    UIDL();

  //Find the message ID
#pragma warning(suppress: 26489)
  auto nFind{std::find(m_msgIDs.begin(), m_msgIDs.end(), pszID)};
  if (nFind != m_msgIDs.end())
#pragma warning(suppress: 26472)
    return static_cast<int>(nFind - m_msgIDs.begin() + 1); //nMsg is 1 based
  return -1;
}

void CPJNPOP3Connection::List()
{
  //Must be connected to perform a "LIST"
#pragma warning(suppress: 26477)
  ATLASSERT(m_bConnected);

  //if we haven't executed the STAT command then do it now
  int nNumberOfMails{m_nNumberOfMails};
  int nTotalMailSize{0};
  if (!m_bStatRetrieved)
    Statistics(nNumberOfMails, nTotalMailSize);

  //Send the LIST command
  static const char* sBuf{"LIST\r\n"};
  try
  {
#pragma warning(suppress: 26472)
    _Send(sBuf, static_cast<int>(strlen(sBuf)));
  }
#ifdef CWSOCKET_MFC_EXTENSIONS
#pragma warning(suppress: 26429 26462)
  catch(CWSocketException* pEx)
  {
#pragma warning(suppress: 26472)
    const DWORD dwError{static_cast<DWORD>(pEx->m_nError)};
    pEx->Delete();
    ThrowPJNPOP3Exception(dwError, FACILITY_WIN32);
  }
#else
#pragma warning(suppress: 26496)
  catch(CWSocketException& e)
  {
    ThrowPJNPOP3Exception(e.m_nError, FACILITY_WIN32);
  }
#endif //#ifdef CWSOCKET_MFC_EXTENSIONS

  //And check the response
  ReadListResponse(nNumberOfMails);
  m_bListRetrieved = true;
}

void CPJNPOP3Connection::UIDL()
{
  //Must be connected to perform a "UIDL"
#pragma warning(suppress: 26477)
  ATLASSERT(m_bConnected);

  //if we haven't executed the STAT command then do it now
  int nNumberOfMails{m_nNumberOfMails};
  int nTotalMailSize{0};
  if (!m_bStatRetrieved)
    Statistics(nNumberOfMails, nTotalMailSize);

  //Send the UIDL command
  static const char* sBuf{"UIDL\r\n"};
  try
  {
#pragma warning(suppress: 26472)
    _Send(sBuf, static_cast<int>(strlen(sBuf)));
  }
#ifdef CWSOCKET_MFC_EXTENSIONS
#pragma warning(suppress: 26429 26462)
  catch(CWSocketException* pEx)
  {
#pragma warning(suppress: 26472)
    const DWORD dwError{static_cast<DWORD>(pEx->m_nError)};
    pEx->Delete();
    ThrowPJNPOP3Exception(dwError, FACILITY_WIN32);
  }
#else
#pragma warning(suppress: 26496)
  catch(CWSocketException& e)
  {
    ThrowPJNPOP3Exception(e.m_nError, FACILITY_WIN32);
  }
#endif //#ifdef CWSOCKET_MFC_EXTENSIONS

  //And check the response
  ReadUIDLResponse(nNumberOfMails);
  m_bUIDLRetrieved = true;
}

void CPJNPOP3Connection::Reset()
{
  //Must be connected to perform a "RSET"
#pragma warning(suppress: 26477)
  ATLASSERT(m_bConnected);

  //Send the RSET command
  static const char* sBuf{"RSET\r\n"};
  try
  {
#pragma warning(suppress: 26472)
    _Send(sBuf, static_cast<int>(strlen(sBuf)));
  }
#ifdef CWSOCKET_MFC_EXTENSIONS
#pragma warning(suppress: 26429 26462)
  catch(CWSocketException* pEx)
  {
#pragma warning(suppress: 26472)
    const DWORD dwError{static_cast<DWORD>(pEx->m_nError)};
    pEx->Delete();
    ThrowPJNPOP3Exception(dwError, FACILITY_WIN32);
  }
#else
#pragma warning(suppress: 26496)
  catch(CWSocketException& e)
  {
    ThrowPJNPOP3Exception(e.m_nError, FACILITY_WIN32);
  }
#endif //#ifdef CWSOCKET_MFC_EXTENSIONS

  //And check the command
  if (!ReadCommandResponse())
    ThrowPJNPOP3Exception(ERROR_UNEXP_NET_ERR, FACILITY_WIN32, m_sLastCommandResponse);
}

void CPJNPOP3Connection::Noop()
{
  //Must be connected to perform a "NOOP"
#pragma warning(suppress: 26477)
  ATLASSERT(m_bConnected);

  //Send the NOOP command
  static const char* sBuf{"NOOP\r\n"};
  try
  {
#pragma warning(suppress: 26472)
    _Send(sBuf, static_cast<int>(strlen(sBuf)));
  }
#ifdef CWSOCKET_MFC_EXTENSIONS
#pragma warning(suppress: 26429 26462)
  catch(CWSocketException* pEx)
  {
#pragma warning(suppress: 26472)
    const DWORD dwError{static_cast<DWORD>(pEx->m_nError)};
    pEx->Delete();
    ThrowPJNPOP3Exception(dwError, FACILITY_WIN32);
  }
#else
#pragma warning(suppress: 26496)
  catch(CWSocketException& e)
  {
    ThrowPJNPOP3Exception(e.m_nError, FACILITY_WIN32);
  }
#endif //#ifdef CWSOCKET_MFC_EXTENSIONS

  //And check the response
  if (!ReadCommandResponse())
    ThrowPJNPOP3Exception(ERROR_UNEXP_NET_ERR, FACILITY_WIN32, m_sLastCommandResponse);
}

void CPJNPOP3Connection::Retrieve(_In_ int nMsg, _Out_ CPJNPOP3Buffer& message)
{
  //Must be connected to retrieve a message
#pragma warning(suppress: 26477)
  ATLASSERT(m_bConnected);

  //work out the size of the message to retrieve
  const DWORD dwSize{GetMessageSize(nMsg)};

  //Send the RETR command
  std::ostringstream ss;
  ss << "RETR " << nMsg << "\r\n";
  const std::string sRequest{ss.str()};
  try
  {
#pragma warning(suppress: 26472)
    _Send(sRequest.c_str(), static_cast<int>(sRequest.length()));
  }
#ifdef CWSOCKET_MFC_EXTENSIONS
#pragma warning(suppress: 26429 26462)
  catch(CWSocketException* pEx)
  {
#pragma warning(suppress: 26472)
    const DWORD dwError{static_cast<DWORD>(pEx->m_nError)};
    pEx->Delete();
    ThrowPJNPOP3Exception(dwError, FACILITY_WIN32);
  }
#else
#pragma warning(suppress: 26496)
  catch(CWSocketException& e)
  {
    ThrowPJNPOP3Exception(e.m_nError, FACILITY_WIN32);
  }
#endif //#ifdef CWSOCKET_MFC_EXTENSIONS

  //And check the command
  ReadReturnResponse(message, dwSize);
}

void CPJNPOP3Connection::GetMessageHeader(_In_ int nMsg, _Out_ CPJNPOP3Buffer& message)
{
  //Must be connected to retrieve a message
#pragma warning(suppress: 26477)
  ATLASSERT(m_bConnected);

  //make sure the message actually exists
  const DWORD dwSize{GetMessageSize(nMsg)};

  //Send the TOP command
  std::ostringstream ss;
  ss << "TOP " << nMsg << " 0\r\n";
  const std::string sRequest(ss.str());
  try
  {
      #pragma warning(suppress: 26472)
      _Send(sRequest.c_str(), static_cast<int>(sRequest.length()));
  }
#ifdef CWSOCKET_MFC_EXTENSIONS
#pragma warning(suppress: 26429 26462)
  catch(CWSocketException* pEx)
  {
#pragma warning(suppress: 26472)
    const DWORD dwError{static_cast<DWORD>(pEx->m_nError)};
    pEx->Delete();
    ThrowPJNPOP3Exception(dwError, FACILITY_WIN32);
  }
#else
#pragma warning(suppress: 26496)
  catch(CWSocketException& e)
  {
    ThrowPJNPOP3Exception(e.m_nError, FACILITY_WIN32);
  }
#endif //#ifdef CWSOCKET_MFC_EXTENSIONS

  //And check the command
  ReadReturnResponse(message, dwSize);
}

bool CPJNPOP3Connection::ReadCommandResponse()
{
  //Delegate to the real implementation
  CPJNPOP3String sResponse;
  return ReadCommandResponse(sResponse);
}

bool CPJNPOP3Connection::ReadCommandResponse(_Inout_ CPJNPOP3String& sResponse)
{
  //Read the response
  CPJNPOP3Buffer response;
  response.reserve(1024);
  bool bSuccess{ReadResponse(response, "\r\n", false)};
  sResponse = response.AsString();

  //determine if the response is an error
  if (bSuccess)
    bSuccess = (response.Find("+OK") == 0);
  return bSuccess;
}

CPJNPOP3Buffer CPJNPOP3Connection::GetBodyOfResponse(_In_ const CPJNPOP3Buffer& response) const
{
  //What will be the return value from this function
  CPJNPOP3Buffer body;

  const int nFind{response.Find("\n")};
  if (nFind != -1)
  {
    const size_t nBodySize{response.size() - nFind - 1};
    if (nBodySize)
#pragma warning(suppress: 26444)
      body.insert(body.end(), response.begin() + nFind + 1, response.end());
  }
  return body;
}

void CPJNPOP3Connection::ReadReturnResponse(_Inout_ CPJNPOP3Buffer& message, _In_ DWORD dwSize)
{
  //Must be connected to perform a "RETR"
#pragma warning(suppress: 26477)
  ATLASSERT(m_bConnected);

  //Retrieve the message body
  CPJNPOP3Buffer response;
#pragma warning(suppress: 26472)
  response.reserve(static_cast<size_t>(dwSize) + 4096);
  const bool bReadResponse{ReadResponse(response, "\r\n.\r\n", true)};
  if (!bReadResponse)
    ThrowPJNPOP3Exception(ERROR_UNEXP_NET_ERR, FACILITY_WIN32, m_sLastCommandResponse);

  //determine if the response is an error
  if (response.Find("+OK") != 0)
    ThrowPJNPOP3Exception(ERROR_UNEXP_NET_ERR, FACILITY_WIN32, m_sLastCommandResponse);

  //remove the first line which contains the +OK from the message and put it into the message
  message = GetBodyOfResponse(response);
  message.resize(message.size() - 5); //Throw away the last "\r\n.\r\n" part of the response
}

void CPJNPOP3Connection::ReadListResponse(_In_ int nNumberOfMails)
{
  //Must be connected to perform a "LIST"
#pragma warning(suppress: 26477)
  ATLASSERT(m_bConnected);

  //retrieve the response
  CPJNPOP3Buffer response;
#pragma warning(suppress: 26472)
  response.reserve((static_cast<size_t>(14) * nNumberOfMails) + 100);
  if (!ReadResponse(response, "\r\n.\r\n", false))
    ThrowPJNPOP3Exception(ERROR_UNEXP_NET_ERR, FACILITY_WIN32, m_sLastCommandResponse);

  //determine if the response is an error
  if (response.Find("+OK") != 0)
    ThrowPJNPOP3Exception(ERROR_UNEXP_NET_ERR, FACILITY_WIN32, m_sLastCommandResponse);

  //Retrieve the message sizes and put them into the m_msgSizes array
  m_msgSizes.clear();
  m_msgSizes.reserve(nNumberOfMails);

  //then parse the LIST response
  CPJNPOP3String sBody{GetBodyOfResponse(response).AsString()};
  if (sBody.empty())
    ThrowPJNPOP3Exception(ERROR_UNEXP_NET_ERR, FACILITY_WIN32, m_sLastCommandResponse);
#pragma warning(suppress: 26446)
  for (int i{0}; (sBody[i] != _T('.')) && (sBody[i] != _T('\0')); i++)
  {
#pragma warning(suppress: 26446)
    if (sBody[i] == _T('\t') || sBody[i] == _T(' '))
    {
#pragma warning(suppress: 26446)
      const TCHAR& chr{sBody[i]};
#pragma warning(suppress: 26489 26486)
      m_msgSizes.push_back(_ttoi(&chr));
    }
  }
}

void CPJNPOP3Connection::ReadUIDLResponse(_In_ int nNumberOfMails)
{
  //Must be connected to perform a "UIDL"
#pragma warning(suppress: 26477)
  ATLASSERT(m_bConnected);

  //retrieve the response
  CPJNPOP3Buffer response;
#pragma warning(suppress: 26472)
  response.reserve((static_cast<size_t>(14) * nNumberOfMails) + 100);
  if (!ReadResponse(response, "\r\n.\r\n", false))
    ThrowPJNPOP3Exception(ERROR_UNEXP_NET_ERR, FACILITY_WIN32, m_sLastCommandResponse);

  //determine if the response is an error
  if (response.Find("+OK") != 0)
    ThrowPJNPOP3Exception(ERROR_UNEXP_NET_ERR, FACILITY_WIN32, m_sLastCommandResponse);

  //Reset the message id array
  m_msgIDs.clear();
  m_msgIDs.reserve(nNumberOfMails);

  //then parse the UIDL response
  CPJNPOP3String sBody{GetBodyOfResponse(response).AsString()};
  if (sBody.empty())
    ThrowPJNPOP3Exception(ERROR_UNEXP_NET_ERR, FACILITY_WIN32, m_sLastCommandResponse);
#pragma warning(suppress: 26446 26429)
  LPTSTR pszSize{sBody.data()};
  for (int iCount=0; iCount<nNumberOfMails; iCount++)
  {
    //Find the next line, before we attempt to parse the current one
    LPTSTR pszBegin{pszSize};
#pragma warning(suppress: 26489)
    while (*pszSize != _T('\n') && *pszSize != _T('\0'))
#pragma warning(suppress: 26481)
      ++pszSize;

    //Convert the EOL to a nullptr so that we can do easier parsing of the message ids
#pragma warning(suppress: 26489)
    if (*pszSize == _T('\n'))
      *pszSize = _T('\0');

    //Find the first whitespace character in the line (which means that we have skipped over the message number)
#pragma warning(suppress: 26430)
    TCHAR* pszID{pszBegin};
#pragma warning(suppress: 26489)
    while (*pszID != _T('\0') && *pszID != _T(' '))
#pragma warning(suppress: 26481)
      ++pszID;

    //Then find the first non-whitespace character in the line (which is the start of the message id)
#pragma warning(suppress: 26489)
    while (*pszID != _T('\0') && *pszID == _T(' '))
#pragma warning(suppress: 26481)
      ++pszID;

    //Then find the first whitespace or EOL character in the line (which is the end of the message id)
#pragma warning(suppress: 26429)
    TCHAR* pszEndID{pszID};
#pragma warning(suppress: 26489)
    while (*pszEndID != _T('\0') && *pszEndID != _T(' ') && *pszEndID != _T('\r') && *pszEndID != _T('\n'))
#pragma warning(suppress: 26481)
      ++pszEndID;
#pragma warning(suppress: 26489)
    *pszEndID = _T('\0');

    //If we could not find the beginning of the ID or the ID is empty, then we were given a bad response
#pragma warning(suppress: 26477 26486 26489)
    if (*pszID == _T('\0') || (_tcslen(pszID) == 0))
      ThrowPJNPOP3Exception(ERROR_UNEXP_NET_ERR, FACILITY_WIN32, m_sLastCommandResponse);

    //Finally form the string from the ID
#pragma warning(suppress: 26477)
    ATLASSUME(pszID != nullptr);
#pragma warning(suppress: 26486)
    m_msgIDs.emplace_back(pszID);
#pragma warning(suppress: 26481)
    pszSize++;
  }
#pragma warning(suppress: 26472)
  if (static_cast<int>(m_msgIDs.size()) != nNumberOfMails) //If we have not got the same amount of elements in m_msgIDs compared to the nNumberOfMails, then we were given a bad response
    ThrowPJNPOP3Exception(ERROR_UNEXP_NET_ERR, FACILITY_WIN32, m_sLastCommandResponse);
}

void CPJNPOP3Connection::ReadStatResponse(_Out_ int& nNumberOfMails, _Out_ int& nTotalMailSize)
{
  //Must be connected to perform a "STAT"
#pragma warning(suppress: 26477)
  ATLASSERT(m_bConnected);

  //retrieve the response
  CPJNPOP3Buffer response;
  response.reserve(128);
  if (!ReadResponse(response, "\r\n", false))
    ThrowPJNPOP3Exception(ERROR_UNEXP_NET_ERR, FACILITY_WIN32, m_sLastCommandResponse);

  //determine if the response is an error
  if (response.Find("+OK") != 0)
    ThrowPJNPOP3Exception(ERROR_UNEXP_NET_ERR, FACILITY_WIN32, m_sLastCommandResponse);

  //Parse out the "Number of Mails" and "Total mail size" values
  CPJNPOP3String sResponse{response.AsString()};
  LPTSTR pszNextToken{nullptr};
#pragma warning(suppress: 26446)
  LPTSTR pszToken{_tcstok_s(sResponse.data(), _T(" \t"), &pszNextToken)};
  if (pszToken == nullptr)
    ThrowPJNPOP3Exception(ERROR_UNEXP_NET_ERR, FACILITY_WIN32, m_sLastCommandResponse);
  pszToken = _tcstok_s(nullptr, _T(" \t"), &pszNextToken);
  if (pszToken == nullptr)
    ThrowPJNPOP3Exception(ERROR_UNEXP_NET_ERR, FACILITY_WIN32, m_sLastCommandResponse);
#pragma warning(suppress: 26477)
  ATLASSUME(pszToken != nullptr);
  nNumberOfMails = _ttoi(pszToken);
  m_nNumberOfMails = nNumberOfMails;
  pszToken = _tcstok_s(nullptr, _T(" \t"), &pszNextToken);
  if (pszToken == nullptr)
    ThrowPJNPOP3Exception(ERROR_UNEXP_NET_ERR, FACILITY_WIN32, m_sLastCommandResponse);
#pragma warning(suppress: 26477)
  ATLASSUME(pszToken != nullptr);
  nTotalMailSize = _ttoi(pszToken);
}

bool CPJNPOP3Connection::ReadResponse(_Inout_ CPJNPOP3Buffer& response, _In_z_ LPCSTR pszTerminator, _In_ bool bCalledForReturnResponse)
{
#ifndef CPJNPOP3_NOSSL
  if (GetSocket() != INVALID_SOCKET)
    return ReadSSLResponse(response, pszTerminator, bCalledForReturnResponse);
  else
#endif //#ifndef CPJNPOP3_NOSSL
    return ReadPlainTextResponse(response, pszTerminator, bCalledForReturnResponse);
}

bool CPJNPOP3Connection::ReadPlainTextResponse(_Inout_ CPJNPOP3Buffer& response, _In_z_ LPCSTR pszTerminator, _In_ bool bCalledForReturnResponse)
{
  //must have been created first
#pragma warning(suppress: 26477)
  ATLASSERT(m_bConnected);

  //Allocate a buffer to read into from the socket
  std::vector<BYTE> readBuffer{m_dwReadBufferSize, std::allocator<BYTE>{}};

  //retrieve the response using until we get the terminator or a timeout occurs
  bool bFoundTerminator{false};
  const size_t nTerminatorLen{strlen(pszTerminator)};
  while (!bFoundTerminator)
  {
    //check the socket for readability
    try
    {
      if (!m_Socket.IsReadable(m_dwTimeout)) //A timeout has occurred so fail the function call
      {
        //Hive away the last command response
        if (bCalledForReturnResponse)
          SetReturnLastCommandResponse(response);
        else
          m_sLastCommandResponse = response.AsString();
        return false;
      }
    }
  #ifdef CWSOCKET_MFC_EXTENSIONS
#pragma warning(suppress: 26429 26462)
    catch(CWSocketException* pEx)
    {
#pragma warning(suppress: 26472)
      const DWORD dwError{static_cast<DWORD>(pEx->m_nError)};
      pEx->Delete();
      ThrowPJNPOP3Exception(dwError, FACILITY_WIN32);
    }
  #else
#pragma warning(suppress: 26496)
    catch(CWSocketException& e)
    {
      ThrowPJNPOP3Exception(e.m_nError, FACILITY_WIN32);
    }
  #endif //#ifdef CWSOCKET_MFC_EXTENSIONS

    //receive the data from the socket
    int nData{0};
    try
    {
      nData = m_Socket.Receive(readBuffer.data(), m_dwReadBufferSize);
    }
  #ifdef CWSOCKET_MFC_EXTENSIONS
#pragma warning(suppress: 26429 26462)
    catch(CWSocketException* pEx)
    {
#pragma warning(suppress: 26472)
      const DWORD dwError{static_cast<DWORD>(pEx->m_nError)};
      pEx->Delete();
      ThrowPJNPOP3Exception(dwError, FACILITY_WIN32);
    }
  #else
#pragma warning(suppress: 26496)
    catch(CWSocketException& e)
    {
      ThrowPJNPOP3Exception(e.m_nError, FACILITY_WIN32);
    }
  #endif //#ifdef CWSOCKET_MFC_EXTENSIONS

    //Handle a graceful disconnect
    if (nData == 0)
      ThrowPJNPOP3Exception(ERROR_GRACEFUL_DISCONNECT, FACILITY_WIN32, m_sLastCommandResponse);

    //Add the data to the complete response
#pragma warning(suppress: 26444 26481 26486)
    response.insert(response.end(), readBuffer.begin(), readBuffer.begin() + nData);

    //Optimize the location where we search for the terminator in the received buffer. This helps 
    //avoid excessive CPU utilization when we receive large responses
#pragma warning(suppress: 26472)
    int nStartSearch{static_cast<int>(response.size() - nData - nTerminatorLen + 1)};
    if (nStartSearch < 0)
      nStartSearch = 0;

    //Check to see if the terminator character(s) have been found
    bFoundTerminator = (response.Find(pszTerminator, nStartSearch) != -1);
  }

  //Hive away the last command response
  if (bCalledForReturnResponse)
    SetReturnLastCommandResponse(response);
  else
    m_sLastCommandResponse = response.AsString();

  return true;
}

#ifndef CPJNPOP3_NOSSL
bool CPJNPOP3Connection::ReadSSLResponse(_Out_ CPJNPOP3Buffer& response, _In_z_ LPCSTR pszTerminator, _In_ bool bCalledForReturnResponse)
{
  //must have been created first
#pragma warning(suppress: 26477)
  ATLASSERT(m_bConnected);

  //Retrieve the response using until we get the terminator or a timeout occurs
  bool bFoundTerminator{false};
  const size_t nTerminatorLen{strlen(pszTerminator)};
  while (!bFoundTerminator)
  {
    //receive the data from the socket
    std::vector<BYTE> message;
    SECURITY_STATUS ss{GetEncryptedMessage(message)};
    if (ss == SEC_I_RENEGOTIATE)
    {
      ss = SSLHandleRenegotiationClient(m_sHostName.c_str());
      if (ss != SEC_E_OK)
        ThrowPJNPOP3Exception(ss);
      continue;
    }
    else if (ss != SEC_E_OK)
      ThrowPJNPOP3Exception(ss);

    //Handle a graceful disconnect
    if (message.size() == 0)
      ThrowPJNPOP3Exception(ERROR_GRACEFUL_DISCONNECT, FACILITY_WIN32, m_sLastCommandResponse);

    //Add the data to the complete response
#pragma warning(suppress: 26444)
    response.insert(response.end(), message.begin(), message.end());

    //Optimize the location where we search for the terminator in the received buffer. This helps 
    //avoid excessive CPU utilization when we receive large responses
#pragma warning(suppress: 26472)
    int nStartSearch{static_cast<int>(response.size() - message.size() - nTerminatorLen + 1)};
    if (nStartSearch < 0)
      nStartSearch = 0;

    //Check to see if the terminator character(s) have been found
    bFoundTerminator = (response.Find(pszTerminator, nStartSearch) != -1);
  }

  //Hive away the last command response
  if (bCalledForReturnResponse)
    SetReturnLastCommandResponse(response);
  else
  {
#pragma warning(suppress: 6001)
    m_sLastCommandResponse = response.AsString();
  }

  return true;
}
#endif //#ifndef CPJNPOP3_NOSSL

void CPJNPOP3Connection::SetReturnLastCommandResponse(_In_ const CPJNPOP3Buffer& response)
{
  //Only copy up to the first EOL if present
  const int nFind{response.Find("\r")};
  if (nFind != -1)
    m_sLastCommandResponse = response.Left(nFind).AsString();
  else
    m_sLastCommandResponse.clear();
}
