/*
Module : PJNPOP3.h
Purpose: Defines the interface for a C++ class framework encapsulation of the POP3 protocol
Created: PJN / 04-05-1998

Copyright (c) 1998 - 2023 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Includes / Defines ///////////////////////////////////////

#pragma once

#if _MSVC_LANG < 201703
#error PJNPOP3 requires a minimum C++ language standard of /std:c++17
#endif //#if _MSVC_LANG < 201703

#ifndef __PJNPOP3_H__
#define __PJNPOP3_H__

#ifndef _EXCEPTION_
#pragma message("To avoid this message, please put exception in your pre compiled header (normally stdafx.h)")
#include <exception>
#endif //#ifndef _EXCEPTION_

#ifndef _STRING_
#pragma message("To avoid this message, please put string in your pre compiled header (normally stdafx.h)")
#include <string>
#endif //#ifndef _STRING_

#ifndef _VECTOR_
#pragma message("To avoid this message, please put vector in your pre compiled header (normally stdafx.h)")
#include <vector>
#endif //#ifndef _VECTOR_

#ifndef _SSTREAM_
#pragma message("To avoid this message, please put sstream in your pre compiled header (normally stdafx.h)")
#include <sstream>
#endif //#ifndef _SSTREAM_

#ifndef _ALGORITHM_
#pragma message("To avoid this message, please put algorithm in your pre compiled header (normally stdafx.h)")
#include <algorithm>
#endif //#ifndef _ALGORITHM_

#ifndef _MEMORY_
#pragma message("To avoid this message, please put memory in your pre compiled header (normally stdafx.h)")
#include <memory>
#endif //#ifndef _MEMORY_

#include "SocMFC.h" //If you get a compilation error about this missing header file, then you need to download my CWSocket wrapper classes from http://www.naughter.com/w3mfc.html

#ifndef CPJNPOP3_NOSSL
#include "SSLWrappers.h" //If you get a compilation error about this missing header file, then you need to download my SSLWrappers code from http://www.naughter.com/sslwrappers.html
#endif //#ifndef CPJNPOP3_NOSSL

#ifndef PJNPOP3_EXT_CLASS
#define PJNPOP3_EXT_CLASS
#endif //#ifndef PJNPOP3_EXT_CLASS

#ifdef _UNICODE
using CPJNPOP3String = std::wstring;
#else
using CPJNPOP3String = std::string;
#endif //#ifdef _UNICODE


//////////////////// Classes //////////////////////////////////////////////////

//The exception class which CPJNPOP3Connection can throw
class PJNPOP3_EXT_CLASS CPJNPOP3Exception : public std::exception
{
public:
//Constructors / Destructors
  CPJNPOP3Exception(_In_ HRESULT hr, _In_ CPJNPOP3String sLastResponse = CPJNPOP3String()) noexcept;
  CPJNPOP3Exception(_In_ DWORD dwError = 0, _In_ DWORD dwFacility = FACILITY_WIN32, _In_ CPJNPOP3String sLastResponse = CPJNPOP3String()) noexcept;

//Methods
  virtual bool GetErrorMessage(_Out_writes_z_(nMaxError) LPTSTR lpszError, _In_ UINT nMaxError, _Out_opt_ PUINT pnHelpContext = nullptr);

//Data members
  HRESULT m_hr;
  CPJNPOP3String m_sLastResponse;
#pragma warning(suppress: 26495)
};


//A simple wrapper for a POP3 response buffer
class CPJNPOP3Buffer : public std::vector<BYTE>
{
public:
//Methods
  _NODISCARD CPJNPOP3String AsString() const;
  _NODISCARD CPJNPOP3Buffer Left(_In_ INT_PTR nCount) const;
  int Find(_In_z_ LPCSTR pszToFind, _In_ int nStart = 0) const;
};


//The main class which encapsulates the POP3 connection
class PJNPOP3_EXT_CLASS CPJNPOP3Connection
#ifndef CPJNPOP3_NOSSL
                                           : protected SSLWrappers::CSocket
#endif //#ifndef CPJNPOP3_NOSSL
{
public:
//typedefs
  enum class AuthenticationMethod
  {
    AUTH_PLAIN = 0, //Username and password are send in the clear to the server (USER / PASS commands)
    AUTH_APOP = 1   //Use APOP support which is a challenge response authentication method (APOP command)
  };

  enum class SSLProtocol
  {
    SSLv2orv3 = 0,
    TLSv1 = 1,
    TLSv1_1 = 2,
    TLSv1_2 = 3,
    //Note enum value which was 4 is defunct
    SSLv2 = 5,
    SSLv3 = 6,
    OSDefault = 7,
    AnyTLSUpTo1_2 = 8,
    AnyTLSUpTo1_3 = 9,
    TLSv1_3 = 10
  };

//Constructors / Destructors
  CPJNPOP3Connection() noexcept;
  CPJNPOP3Connection(const CPJNPOP3Connection&) = delete;
  CPJNPOP3Connection(CPJNPOP3Connection&&) = delete;
#ifndef CPJNPOP3_NOSSL
  ~CPJNPOP3Connection(); //NOLINT(modernize-use-override)
#else
  virtual ~CPJNPOP3Connection();
#endif //#ifndef CPJNPOP3_NOSSL

//Methods
  CPJNPOP3Connection& operator=(const CPJNPOP3Connection&) = delete;
  CPJNPOP3Connection& operator=(CPJNPOP3Connection&&) = delete;
#ifndef CPJNPOP3_NOSSL
  void Connect(_In_z_ LPCTSTR pszHostName, _In_z_ LPCTSTR pszUsername, _In_z_ LPCTSTR pszPassword, _In_ int nPort = 110, _In_ AuthenticationMethod am = AuthenticationMethod::AUTH_PLAIN, _In_ bool bSSL = false);
#else
  void Connect(_In_z_ LPCTSTR pszHostName, _In_z_ LPCTSTR pszUsername, _In_z_ LPCTSTR pszPassword, _In_ int nPort = 110, _In_ AuthenticationMethod am = AuthenticationMethod::AUTH_PLAIN);
#endif //#ifndef CPJNPOP3_NOSSL
  void Disconnect(_In_ bool bGracefully = true);
  _NODISCARD bool IsConnected() const noexcept { return m_bConnected; };
  _NODISCARD CPJNPOP3String GetLastCommandResponse() const { return m_sLastCommandResponse; };
  _NODISCARD DWORD GetTimeout() const noexcept { return m_dwTimeout; };
  void SetTimeout(_In_ DWORD dwTimeout) noexcept { m_dwTimeout = dwTimeout; };
  void Statistics(_Out_ int& nNumberOfMails, _Out_ int& nTotalMailSize);
  void Delete(_In_ int nMsg);
  DWORD GetMessageSize(_In_ int nMsg);
  CPJNPOP3String GetMessageID(_In_ int nMsg);
  int FindMessageID(_In_z_ LPCTSTR pszID);
  void Retrieve(_In_ int nMsg, _Out_ CPJNPOP3Buffer& message);
  void GetMessageHeader(_In_ int nMsg, _Out_ CPJNPOP3Buffer& message);
  void Reset();
  void UIDL();
  void Noop();
  void List();
  void SetReadBufferSize(_In_ DWORD dwReadBufferSize) noexcept { m_dwReadBufferSize = dwReadBufferSize; };
  _NODISCARD DWORD GetReadBufferSize() const noexcept { return m_dwReadBufferSize; };
  void SetSSLProtocol(_In_ SSLProtocol protocol) noexcept { m_SSLProtocol = protocol; }
  _NODISCARD SSLProtocol GetSSLProtocol() const noexcept { return m_SSLProtocol; }
  void SetBindAddress(_In_ const CPJNPOP3String& sBindAddress) { m_sBindAddress = sBindAddress; };
  _NODISCARD CPJNPOP3String SetBindAddress() const { return m_sBindAddress; };
  CWSocket& GetPlainTextSocket() noexcept { return m_Socket; };

//Static methods
  static void ThrowPJNPOP3Exception(_In_ DWORD dwError = 0, _In_ DWORD Facility = FACILITY_WIN32, _In_ const CPJNPOP3String& sLastResponse = _T(""));
  static void ThrowPJNPOP3Exception(_In_ HRESULT hr, _In_ const CPJNPOP3String& sLastResponse = _T(""));

protected:
  void _Send(_In_reads_bytes_(nBuf) const void* pBuffer, _In_ int nBuf);
  virtual void ReadStatResponse(_Out_ int& nNumberOfMails, _Out_ int& nTotalMailSize);
  bool ReadCommandResponse();
  bool ReadCommandResponse(_Inout_ CPJNPOP3String& sResponse);
  virtual void ReadListResponse(_In_ int nNumberOfMails);
  virtual void ReadUIDLResponse(_In_ int nNumberOfMails);
  virtual void ReadReturnResponse(_Inout_ CPJNPOP3Buffer& message, _In_ DWORD dwSize);
  bool ReadResponse(_Inout_ CPJNPOP3Buffer& response, _In_z_ LPCSTR pszTerminator, _In_ bool bCalledForReturnResponse);
  bool ReadPlainTextResponse(_Inout_ CPJNPOP3Buffer& response, _In_z_ LPCSTR pszTerminator, _In_ bool bCalledForReturnResponse);
  _NODISCARD virtual CPJNPOP3Buffer GetBodyOfResponse(_In_ const CPJNPOP3Buffer& response) const;
#ifndef CPJNPOP3_NOSSL
  bool ReadSSLResponse(_Out_ CPJNPOP3Buffer& response, _In_z_ LPCSTR pszTerminator, _In_ bool bCalledForReturnResponse);
  virtual void CreateSSLSocket();
#endif //#ifndef CPJNPOP3_NOSSL
  virtual void SetReturnLastCommandResponse(_In_ const CPJNPOP3Buffer& response);

//Member variables
#ifndef CPJNPOP3_NOSSL
  std::unique_ptr<SSLWrappers::CCachedCredentials> m_SSLCredentials; //SSL credentials
#endif //#ifndef CPJNPOP3_NOSSL
  bool m_bSSL; //Are we connecting using SSL?
  CWSocket m_Socket; //The socket connection to the POP3 server (if not using SSL)
  bool m_bConnected; //Are we currently connected to the server
  CPJNPOP3String m_sLastCommandResponse; //The full last response the server sent us
  DWORD m_dwTimeout;
  CPJNPOP3String m_sBindAddress;
  DWORD m_dwReadBufferSize;
  int m_nNumberOfMails;
  bool m_bListRetrieved;
  bool m_bStatRetrieved;
  bool m_bUIDLRetrieved;
  std::vector<DWORD> m_msgSizes;
  std::vector<CPJNPOP3String> m_msgIDs;
  SSLProtocol m_SSLProtocol; //What flavour of TLS/SSL should we speak
  CPJNPOP3String m_sHostName; //The host name we are connecting to
};

#endif //#ifndef __PJNPOP3_H__
